package src.hw8convex;

public class BusModelConvex implements MathFunction {
	private double[] x;
	private double[] y;
	private int n;						// Number of unknowns
	private double[] upBound;
	private double[] lowBound;
	private double p  = 3.59;   // Density
	private double jj  = 0.05;   // Walk access speed
	private double k  = 0.4;    // Wait-hdwy ratio
	private double c  = 50;     // Bus cost
	private double v  = 0.167;  // Bus speed
	private double a0 = 0.41;   // Demand coefficients
	private double a2 = 0.0081;
	private double a4 = 0.0014;
	private double T= 60;   	// Time period of analysis (min)
	private double X= 4;		// Width of analysis area (mi)
	private double Y= 6; 		// Length of analysis area (mi)

	public BusModelConvex() {
		n= 2;
		x= new double[n];
		upBound= new double[n];
		upBound[0]= 1.0;				// Route spacing
		upBound[1]= 200.0;				// Fare
		lowBound= new double[n];		// Lower bounds are zero
		y= new double[n];
	}

	public double getRteSpacing() {
		return Math.pow((64*jj*jj*k*a4*c/(p*v*a0*a2)), 1.0/3.0);
	}

	public double getHeadway() {
		return Math.pow((a4*c/(p*a0*a2*k*k*jj*v)), 1.0/3.0);
	}

	public double getFare() {
		return (0.5*a0/a4 - Math.pow((k*c*a2*a2)/(jj*a4*a4*p*v*a0), 1.0/3.0));
	}

	public double func(double[] x) {
		double g= x[0];
		double f= x[1];
		double h= 0.25*g/(jj*k);
		double cost= 2*X*T*c*Y/(g*h*v);
		double revenue= T*p*X*Y*f*(a0-a2*(k*h+0.25*g/jj)-a4*f);
		return (revenue - cost);		
	}

	public double[] gradient() {
		double[] d= new double[x.length];
		d[0]= -0.25*p*x[1]*a2/jj + 8*jj*k*c/(v*x[0]*x[0]*x[0]);
		d[1]= a0 - 0.5*a2*x[0]/jj - 2*a4*x[1];
		return d;
	}

	public double derivative(double alpha) {
		double k1= p*a0;
		double k2= -0.5*p*a2/jj;
		double k3= -p*a4;
		double k4= -8*jj*k*c/v;
		double d1= 2*alpha* (k2*(y[0]-x[0])*(y[1]-x[1]) + k3*(y[1]-x[1])*(y[1]-x[1]));
		double d2= k1*(y[1]-x[1]) + k2*x[1]*(y[0]-x[0]) + k2*x[0]*(y[1]-x[1]) + 2*k3*x[1]*(y[1]-x[1]);
		double d3= -k4* (2*(y[0]-x[0]) + 2*alpha*(y[0]-x[0])*(y[0]-x[0]));
		double d4= 2*alpha*(y[0]-x[0]) + alpha*alpha*(y[0]-x[0])* (y[0]-x[0]);
		double deriv;
		if (y[0]-x[0]> 0)
			deriv= d1 + d2 + d3/(d4*d4);
		else
			deriv= d1 + d2;
		return deriv;
	}

	private double lineSearch(MathFunction d, double a, double b) {	// Change signs from other versions since this is looking for the maximum, not minimum
		final double TOLERANCE= 1E-8;		// Square root of machine precision
		final double MAX_ITERATIONS= 1000;
		double m;							// Midpoint
		int counter= 0;
		for (m= (a+b)/2.0; Math.abs(a-b) > TOLERANCE; m= (a+b)/2.0) {
			counter++;
			if  (d.derivative(m)> 0.0)    // If derivative positive, 
				a= m;       // Use left subinterval
			else
				b= m;       // Use right subinterval
			if (counter > MAX_ITERATIONS)
				break;
		}
		return m;
	}

	private double[] linearProgram(double[] lpCoeff) {	// Maximization
		for (int i= 0; i < n; i++) {
			if (lpCoeff[i]< 0 )
				y[i]= lowBound[i];
			else 
				y[i]= upBound[i];
		}
		return y;
	}

	private void printDirection(double a) {
		System.out.print("Direction  y: \t");
		for (int i= 0; i < n; i++)
			System.out.printf("%5.2f ",y[i]);
		System.out.printf("               %5.3f", a);
		System.out.println();	
	}  

	private double move(double alpha) {
		double sumX= 0.0;
		double sumRootMeanDiffX= 0.0;
		for (int i= 0; i < n; i++) {
			double variableChange= alpha*(y[i] - x[i]);
			x[i]+= variableChange;
			sumX+= x[i];
			sumRootMeanDiffX+= variableChange*variableChange;
		}
		// Compute convergence criterion here, since we have current and previous x
		return Math.sqrt(sumRootMeanDiffX)/sumX;
	}

	private void printMove() {
		System.out.print("Move       x: \t");
		for (int i= 0; i < n; i++)
			System.out.printf("%5.2f ",x[i]);
		System.out.printf("   %6.3f ", func(x));
		System.out.println("\n");	
	}  
	
	private void printAlpha() {
		for (double alpha= 0.0; alpha <= 1.0; alpha+= 0.01)
			System.out.println(alpha+" "+derivative(alpha));
	}
	
	public void plot() {
		for (double f= 0; f <= 200; f+=10) { 
			System.out.println("Fare: "+ f);
			for (double g= 0.6; g <= 1.0; g += 0.1) {
				x[0]= g;
				x[1]= f;
				System.out.printf("%5.2f %8.1f ",g,func(x));
			}
			System.out.println();
		}
	}

	public void maximize() {
		// Initialization in constructor. Need good first guesses, or method cannot start
		x[0]= Math.min(getRteSpacing(), upBound[0]);
		x[1]= Math.min(getFare(), upBound[1]);
		printMove();
		final double CRITERION= 0.001;		// Convergence criterion: 0.1% change in x (root mean square)
		final int MAX_ITERATIONS= 100;		// Should be set higher in real code
		double convergence;
		int iterations= 0;
		do {
			double[] lpCoeff= gradient();
			y= linearProgram(lpCoeff);		// Constraints are instance variables, never change
//			printAlpha();
			double alpha= lineSearch(this, 0.0, 1.0);	// Alpha between 0 and 1
			printDirection(alpha);
			convergence= move(alpha);
			printMove();
			// Step 5: Check convergence
			iterations++;
		} while (convergence > CRITERION && iterations < MAX_ITERATIONS);
	}

	public static void main(String[] args) {
		BusModelConvex g= new BusModelConvex();
//		g.plot();
		g.maximize();
	}   
}